// PipeInput.cp

#ifndef PipeInput_h
#include "PipeInput.h"
#endif
#ifndef TeeBase_h
#include "TeeBase.h"
#endif
#ifndef MinMax_h
#include "MinMax.h"
#endif

PipeInput::PipeInput( Data theRing, TeeBase& thePipe )
  : buffer( theRing ),
	 ring( theRing ),
	 pipe( thePipe ),
	 acceptLength( 0 ),
	 sequence( *this )
  {
  }

void PipeInput::Reset()
  {
	buffer.Reset( ring );
	acceptLength = 0;
	entering = PipePacket();
	exiting = PipePacket();
  }

void PipeInput::PushIn( PipePacket more )
  {
	entering += more;
	sleep.Wake();
  }

void PipeInput::PushOut()
  {
	Assert( pipe.AuditTotalCorrect() );
	Assert( pipe.OutputAuditsMatch() );
	Assert( pipe.SegmentEndsMatch() );

	if ( !exiting.IsZero() )
	  {
		pipe.PushToOutputs( exiting );
		exiting = PipePacket();
	  }

	Assert( pipe.AuditTotalCorrect() );
	Assert( pipe.OutputAuditsMatch() );
	Assert( pipe.SegmentEndsMatch() );
  }

void PipeInput::MoveBuffer()
  {
	Assert( pipe.AuditTotalCorrect() );
	Assert( pipe.OutputAuditsMatch() );
	Assert( pipe.SegmentEndsMatch() );

	if ( buffer.UsedLength() > 0 )
	  {
		exiting += PipePacket( buffer.UsedLength() );
		buffer.Reset( buffer.Unused() );
	  }
	
	Assert( pipe.AuditTotalCorrect() );
	Assert( pipe.OutputAuditsMatch() );
	Assert( pipe.SegmentEndsMatch() );

	if ( entering.DataSize() > 0 )
	  {
		uint8 *start = ( buffer.UnusedStart() < ring.End() )
							? buffer.UnusedStart()
							: ring.Start();
		
		PipePacket accepted( entering );
		accepted.LimitData( ring.End() - start - buffer.UnusedLength() );

		entering -= accepted;
		buffer.Reset( Data( start,
								  start
								  	+ buffer.UnusedLength()
								  	+ accepted.DataSize() ) );
	  }

	Assert( pipe.AuditTotalCorrect() );
	Assert( pipe.OutputAuditsMatch() );
	Assert( pipe.SegmentEndsMatch() );
  }

TaskStep<PipeInput> PipeInput::DoPass( bool, DeferredTaskTime )
  {
	MoveBuffer();
	PushOut();
	return Step( 0 );
  }

TaskStep<PipeInput> PipeInput::TryAccept( bool dying, DeferredTaskTime )
  {
	MoveBuffer();
	PushOut();
	
	if ( !dying && buffer.UnusedLength() < acceptLength )
		return Step( sleep(), &PipeInput::TryAccept );
	
	return Step( 0 );
  }

TaskStep<PipeInput> PipeInput::StartFlush( bool dying, DeferredTaskTime )
  {
	MoveBuffer();
	exiting += PipePacket::flush;
	PushOut();
	
	if ( dying )
		return Step( 0 );
	
	return Step( sleep(), &PipeInput::FinishFlush );
  }

TaskStep<PipeInput> PipeInput::FinishFlush( bool dying, DeferredTaskTime )
  {
	if ( !dying && !entering.Flushing() )
		return Step( sleep(), &PipeInput::FinishFlush );
	
	Assert( Audit() == RingLength() );
	
	MoveBuffer();
	
	return Step( 0 );
  }

uint32 PipeInput::PotentialLength() const
  {
	if ( buffer.UnusedStart() == ring.End() )
		return RingLength();
	
	return ring.End() - buffer.UnusedStart();
  }

uint32 PipeInput::LargestAcceptSize() const
  {
	return Min( RingLength() / 2, PotentialLength() );
  }

uint32 PipeInput::ReasonableAcceptSize() const
  {
	return Min( RingLength() / 4, PotentialLength() );
  }

uint32 PipeInput::ReasonablePassSize() const
  {
	return Min( RingLength() / 4, buffer.UnusedLength() );
  }

Task *PipeInput::Pass()
  {
	if ( buffer.UsedLength() == 0 )
		return 0;
	
	if ( DeferredTaskTime::IsNow() )
	  {
		MoveBuffer();
		PushOut();
		return 0;
	  }
	
	return sequence.Start( Step( 0, &PipeInput::DoPass ) );
  }
		
Task *PipeInput::Accept( uint32 requiredAmount )
  {
	Assert( requiredAmount <= LargestAcceptSize() );

	if ( DeferredTaskTime::IsNow() )
	  {
		MoveBuffer();
		PushOut();
		if ( buffer.UnusedLength() >= requiredAmount )
			return 0;
	  }
	
	acceptLength = requiredAmount;
	return sequence.Start( Step( 0, &PipeInput::TryAccept ) );
  }

Task *PipeInput::Flush()
  {
	return sequence.Start( Step( 0, &PipeInput::StartFlush ) );
  }

uint32 PipeInput::Audit() const
  {
	return entering.DataSize()
				+ buffer.UsedLength()
				+ buffer.UnusedLength()
				+ exiting.DataSize();
  }

const uint8 *PipeInput::ExitingStart() const
  {
	uint32 position = ring.Length()
							+ ( buffer.UsedStart() - ring.Start() )
							- exiting.DataSize();
	
	return ring.Start() + position % ring.Length();
  }

const uint8 *PipeInput::EnteringEnd() const
  {
	uint32 position = ( buffer.UnusedEnd() - ring.Start() )
							+ entering.DataSize();
	
	return ring.Start() + position % ring.Length();
  }
