// PipeOutput.cp

#ifndef PipeOutput_h
#include "PipeOutput.h"
#endif
#ifndef TeeBase_h
#include "TeeBase.h"
#endif
#ifndef MinMax_h
#include "MinMax.h"
#endif

PipeOutput::PipeOutput()
  : buffer( ConstData() ),
	 pipe( 0 ),
	 acceptLength( 0 ),
	 flushing( false ),
	 sequence( *this )
  {
  }

void PipeOutput::Initialize( ConstData theRing, TeeBase& thePipe )
  {
	Assert( pipe == 0 );
	buffer.Reset( ConstData( theRing.Start(), theRing.Start() ) );
	ring = theRing;
	pipe = &thePipe;
  }

void PipeOutput::Reset()
  {
	Assert( pipe != 0 );
	buffer.Reset( ConstData( ring.Start(), ring.Start() ) );
	acceptLength = 0;
	flushing = false;
	entering = PipePacket();
	exiting = PipePacket();
  }

void PipeOutput::PushIn( PipePacket more )
  {
	entering += more;
	sleep.Wake();
  }

void PipeOutput::MoveBuffer()
  {
	Assert( pipe->AuditTotalCorrect() );
	Assert( pipe->OutputAuditsMatch() );
	Assert( pipe->SegmentEndsMatch() );

	if ( buffer.UsedLength() > 0 )
	  {
		exiting += PipePacket( buffer.UsedLength() );
		buffer.Reset( buffer.Unused() );
	  }

	Assert( pipe->AuditTotalCorrect() );
	Assert( pipe->OutputAuditsMatch() );
	Assert( pipe->SegmentEndsMatch() );
	
	if ( !entering.IsZero() )
	  {
		const uint8 *start = ( buffer.UnusedStart() < ring.End() )
									? buffer.UnusedStart()
									: ring.Start();
		
		PipePacket accepted( entering );
		accepted.LimitData( ring.End() - start - buffer.UnusedLength() );
		
		entering -= accepted;
		buffer.Reset( ConstData( start,
										 start
										 	+ buffer.UnusedLength()
										 	+ accepted.DataSize() ) );
		flushing = accepted.Flushing();
	  }

	Assert( pipe->AuditTotalCorrect() );
	Assert( pipe->OutputAuditsMatch() );
	Assert( pipe->SegmentEndsMatch() );
  }
	
TaskStep<PipeOutput> PipeOutput::DoPass( bool, DeferredTaskTime )
  {
	Assert( pipe != 0 );
	Assert( buffer.UsedLength() > 0 );
	
	MoveBuffer();
	pipe->PushToInput();
	
	return Step( 0 );
  }

TaskStep<PipeOutput> PipeOutput::TryAccept( bool dying, DeferredTaskTime )
  {
	Assert( pipe != 0 );
	
	MoveBuffer();
	if ( !exiting.IsZero() )
		pipe->PushToInput();
	
	if ( !dying && buffer.UnusedLength() < acceptLength )
		return Step( sleep(), &PipeOutput::TryAccept );
	
	return Step( 0 );
  }

TaskStep<PipeOutput> PipeOutput::DoAcknowledgeFlush( bool, DeferredTaskTime )
  {
	Assert( pipe != 0 );
	Assert( flushing );
	Assert( buffer.Exhausted() );
	
	MoveBuffer();
	flushing = false;
	exiting += PipePacket::flush;
	pipe->PushToInput();
	
	return Step( 0 );
  }

uint32 PipeOutput::PotentialLength() const
  {
	if ( buffer.UnusedStart() == ring.End() )
		return RingLength();
	
	return ring.End() - buffer.UnusedStart();
  }

uint32 PipeOutput::LargestAcceptSize() const
  {
	return Min( RingLength() / 2, PotentialLength() );
  }

uint32 PipeOutput::ReasonableAcceptSize() const
  {
	return Min( RingLength() / 4, PotentialLength() );
  }

uint32 PipeOutput::ReasonablePassSize() const
  {
	if ( Flushing() )
		return buffer.UnusedLength();
	
	return Min( RingLength() / 4, buffer.UnusedLength() );
  }

Task *PipeOutput::Pass()
  {
	Assert( pipe != 0 );

	if ( buffer.UsedLength() == 0 )
		return 0;
	
	if ( DeferredTaskTime::IsNow() )
	  {
		MoveBuffer();
		pipe->PushToInput();
		return 0;
	  }
	
	return sequence.Start( Step( 0, &PipeOutput::DoPass ) );
  }

Task *PipeOutput::Accept( uint32 requiredAmount )
  {
	Assert( pipe != 0 );
	Assert( requiredAmount <= LargestAcceptSize() );
	
	Assert( !flushing || requiredAmount <= buffer.UnusedLength() );

	if ( DeferredTaskTime::IsNow() )
	  {
		MoveBuffer();
		if ( !exiting.IsZero() )
			pipe->PushToInput();
	
		if ( buffer.UnusedLength() >= requiredAmount )
			return 0;
	  }
	
	acceptLength = requiredAmount;
	return sequence.Start( Step( 0, &PipeOutput::TryAccept ) );
  }

Task *PipeOutput::AcknowledgeFlush()
  {
	Assert( pipe != 0 );
	Assert( buffer.Exhausted() );
	
	return sequence.Start( Step( 0, &PipeOutput::DoAcknowledgeFlush ) );
  }

uint32 PipeOutput::Audit() const
  {
	return entering.DataSize()
				+ buffer.UsedLength()
				+ buffer.UnusedLength()
				+ exiting.DataSize();
  }

const uint8 *PipeOutput::ExitingStart() const
  {
	uint32 position = ring.Length()
							+ ( buffer.UsedStart() - ring.Start() )
							- exiting.DataSize();
	
	return ring.Start() + position % ring.Length();
  }

const uint8 *PipeOutput::EnteringEnd() const
  {
	uint32 position = ( buffer.UnusedEnd() - ring.Start() )
							+ entering.DataSize();
	
	return ring.Start() + position % ring.Length();
  }
